{-# LANGUAGE MultiParamTypeClasses  #-}

{-| Module  : FiniteCategories
Description : A comma category is a category where objects are morphisms of another category /C/ and morphisms are commutative squares in /C/.
Copyright   : Guillaume Sabbagh 2021
License     : GPL-3
Maintainer  : guillaumesabbagh@protonmail.com
Stability   : experimental
Portability : portable

A comma category is a category where objects are morphisms of another category /C/ and morphisms are commutative squares in /C/.

For example, see Examples.ExampleCommaCategory.*
-}

module CommaCategory.CommaCategory
(
    CommaObject(..),
    CommaMorphism(..),
    CommaCategory(..),
    mkSliceCategory,
    mkCosliceCategory,
    mkArrowCategory,
)
where
    import FiniteCategory.FiniteCategory
    import Diagram.Diagram
    import UsualCategories.One
    import IO.PrettyPrint
    import Utils.AssociationList
    
    -- | A `CommaObject` in the `CommaCategory` (/T/|/S/) is a triplet \</e/,/d/,/f/\> where @f : /T/(e) -> /S/(d)@.
    --
    -- See /Categories for the working mathematician/, Saunders Mac Lane, P.46.
    data CommaObject o1 o2 m3 = CommaObject {indexSrc :: o1 -- ^ /e/
                                        , indexTgt :: o2 -- ^ /d/
                                        , arrow :: m3} -- ^ /f/
                                        deriving (Eq, Show)
    
    instance (PrettyPrintable o1, PrettyPrintable o2, PrettyPrintable m3) =>
             PrettyPrintable (CommaObject o1 o2 m3) where
        pprint CommaObject{indexSrc=e, indexTgt=d, arrow=f} = "<"++pprint e++", "++pprint d++", "++pprint f++">"
    
    -- | A `CommaMorphism` in the `CommaCategory` (/T/|/S/) is a couple \</k/,/h/\>.
    --
    -- See /Categories for the working mathematician/, Saunders Mac Lane, P.46.
    data CommaMorphism o1 o2 m1 m2 m3 = CommaMorphism {srcCM :: (CommaObject o1 o2 m3) -- ^ The source `CommaObject`
                                                     , tgtCM :: (CommaObject o1 o2 m3) -- ^ The target `CommaObject`
                                                     , indexAr1 :: m1 -- ^ /k/
                                                     , indexAr2 :: m2} -- ^ /h/
                                                     deriving (Eq, Show)
                                                     
    instance (PrettyPrintable m1, PrettyPrintable m2) =>
              PrettyPrintable (CommaMorphism o1 o2 m1 m2 m3) where
        pprint CommaMorphism{srcCM=_, tgtCM =_, indexAr1=k, indexAr2=h} = "<"++pprint k++", "++pprint h++">"
    
    instance (Morphism m1 o1, Morphism m2 o2, Eq o1, Eq o2, Eq m3) => Morphism (CommaMorphism o1 o2 m1 m2 m3) (CommaObject o1 o2 m3) where
        (@) CommaMorphism{srcCM=s2,tgtCM=t2,indexAr1=k2,indexAr2=h2} CommaMorphism{srcCM=s1,tgtCM=t1,indexAr1=k1,indexAr2=h1}
            | t1 /= s2 = error "Illegal composition of CommaMorphism."
            | otherwise = CommaMorphism {srcCM=s1,tgtCM=t2,indexAr1=k2 @ k1,indexAr2=h2 @ h1}
        source = srcCM
        target = tgtCM
    
    -- | A `CommaCategory` is a couple (/T/|/S/) with /T/ and /S/ two diagrams with the same target.
    --
    -- See /Categories for the working mathematician/, Saunders Mac Lane, P.46.
    data CommaCategory c1 m1 o1 c2 m2 o2 c3 m3 o3 = CommaCategory {leftDiag :: Diagram c1 m1 o1 c3 m3 o3 -- ^ /T/
                                                                , rightDiag :: Diagram c2 m2 o2 c3 m3 o3} -- ^ /S/
                                                    deriving (Eq, Show)
        
    instance (FiniteCategory c1 m1 o1, Morphism m1 o1, Eq m1, Eq o1,
              FiniteCategory c2 m2 o2, Morphism m2 o2, Eq m2, Eq o2,
              FiniteCategory c3 m3 o3, Morphism m3 o3, Eq m3) => FiniteCategory (CommaCategory c1 m1 o1 c2 m2 o2 c3 m3 o3) (CommaMorphism o1 o2 m1 m2 m3) (CommaObject o1 o2 m3) where
        ob CommaCategory {leftDiag = t, rightDiag = s} = [CommaObject{indexSrc=e,indexTgt=d,arrow=f}| e <- (ob (src t)), d <- (ob (src s)), f <- (ar (tgt t) ((omap t) !-! (e)) ((omap s) !-! (d)))]
        identity CommaCategory{leftDiag = t, rightDiag = s} obj@CommaObject{indexSrc=e,indexTgt=d,arrow=f} 
            = CommaMorphism{srcCM=obj, tgtCM=obj, indexAr1=(identity (src t) e), indexAr2=(identity (src s) d)}
        ar CommaCategory{leftDiag = t, rightDiag = s} obj1@CommaObject{indexSrc=e1,indexTgt=d1,arrow=f1} obj2@CommaObject{indexSrc=e2,indexTgt=d2,arrow=f2}
            = [CommaMorphism{srcCM=obj1,tgtCM=obj2,indexAr1=k,indexAr2=h}| k <- ar (src t) e1 e2, h <- ar (src s) d1 d2, f2 @ ((mmap t) !-! (k)) == ((mmap s) !-! (h)) @ f1]
                    
    instance (FiniteCategory c1 m1 o1, Morphism m1 o1, Eq m1, Eq o1,
              FiniteCategory c2 m2 o2, Morphism m2 o2, Eq m2, Eq o2,
              FiniteCategory c3 m3 o3, Morphism m3 o3, Eq m3) => GeneratedFiniteCategory (CommaCategory c1 m1 o1 c2 m2 o2 c3 m3 o3) (CommaMorphism o1 o2 m1 m2 m3) (CommaObject o1 o2 m3) where
        genAr = defaultGenAr
        decompose = defaultDecompose

    instance (PrettyPrintable c1, PrettyPrintable m1, PrettyPrintable o1,
              PrettyPrintable c2, PrettyPrintable m2, PrettyPrintable o2,
              PrettyPrintable c3, PrettyPrintable m3, PrettyPrintable o3,
              FiniteCategory c1 m1 o1, FiniteCategory c2 m2 o2, FiniteCategory c3 m3 o3,
              Morphism m1 o1, Morphism m2 o2, Morphism m3 o3,
              Eq m1, Eq o1, Eq m2, Eq o2) =>
              PrettyPrintable (CommaCategory c1 m1 o1 c2 m2 o2 c3 m3 o3) where
        pprint CommaCategory{leftDiag=t, rightDiag=s} = "("++pprint t++"|"++pprint s++")"

    -- | Constructs the slice category of a category /C/ over an object /o/.
    --
    -- Returns Nothing if the object is not in the category.
    mkSliceCategory :: (FiniteCategory c m o, Morphism m o, Eq o) => c -> o -> Maybe (CommaCategory c m o One One One c m o)
    mkSliceCategory c o = mkSelect1 c o >>= (\x -> Just CommaCategory{leftDiag=mkIdentityDiagram c, rightDiag=x})
        
    -- | Constructs the coslice category of a category /C/ under an object /o/.
    --
    -- Returns Nothing if the object is not in the category.
    mkCosliceCategory :: (FiniteCategory c m o, Morphism m o, Eq o) => c -> o -> Maybe (CommaCategory One One One c m o c m o)
    mkCosliceCategory c o = mkSelect1 c o >>= (\x -> Just CommaCategory{rightDiag=mkIdentityDiagram c, leftDiag=x})
        
    -- | Constructs the arrow category of a category /C/.
    mkArrowCategory :: (FiniteCategory c m o, Morphism m o, Eq o) => c -> CommaCategory c m o c m o c m o
    mkArrowCategory c = CommaCategory{leftDiag=mkIdentityDiagram c, rightDiag=mkIdentityDiagram c}