{-# LANGUAGE MultiParamTypeClasses  #-}

{-| Module  : FiniteCategories
Description : Product category of two categories.
Copyright   : Guillaume Sabbagh 2021
License     : GPL-3
Maintainer  : guillaumesabbagh@protonmail.com
Stability   : experimental
Portability : portable

Product category of two categories.
-}
module ProductCategory.ProductCategory
(
firstObject,
firstMorphism,
firstCategory,
secondObject,
secondMorphism,
secondCategory,
ProductObject(..),
ProductMorphism(..),
ProductCategory(..),
)
where
    import FiniteCategory.FiniteCategory
    import Utils.CartesianProduct
    import IO.PrettyPrint
    
    -- | Object in a product category.
    data ProductObject o1 o2 = ProductObject o1 o2 deriving (Eq, Show, Ord)
    
    -- | Returns the first object of a product object.
    firstObject :: ProductObject o1 o2 -> o1
    firstObject (ProductObject o1 _) = o1
    
    -- | Returns the second object of a product object.
    secondObject :: ProductObject o1 o2 -> o2
    secondObject (ProductObject _ o2) = o2
    
    -- | Morphism in a product category.
    data ProductMorphism m1 o1 m2 o2 = ProductMorphism m1 m2 deriving (Eq, Show, Ord)
    
    -- | Returns the first morphism of a product morphism.
    firstMorphism :: ProductMorphism m1 o1 m2 o2 -> m1
    firstMorphism (ProductMorphism m1 _) = m1
    
    -- | Returns the second morphism of a product morphism.
    secondMorphism :: ProductMorphism m1 o1 m2 o2 -> m2
    secondMorphism (ProductMorphism _ m2) = m2
    
    -- | Product category of two categories.
    data ProductCategory c1 m1 o1 c2 m2 o2 = ProductCategory c1 c2 deriving (Eq, Show, Ord)

    -- | Returns the first category of a product category.
    firstCategory :: ProductCategory c1 m1 o1 c2 m2 o2 -> c1
    firstCategory (ProductCategory c1 _) = c1
    
    -- | Returns the second category of a product category.
    secondCategory :: ProductCategory c1 m1 o1 c2 m2 o2 -> c2
    secondCategory (ProductCategory _ c2) = c2

    instance (Morphism m1 o1, Morphism m2 o2) => Morphism (ProductMorphism m1 o1 m2 o2) (ProductObject o1 o2) where
        source (ProductMorphism m1 m2) = ProductObject (source m1) (source m2)
        target (ProductMorphism m1 m2) = ProductObject (target m1) (target m2)
        (ProductMorphism g1 g2) @ (ProductMorphism f1 f2) = ProductMorphism (g1 @ f1) (g2 @ f2)
                                                
    instance (FiniteCategory c1 m1 o1, Morphism m1 o1,
              FiniteCategory c2 m2 o2, Morphism m2 o2) =>
              FiniteCategory (ProductCategory c1 m1 o1 c2 m2 o2) (ProductMorphism m1 o1 m2 o2) (ProductObject o1 o2) where
        ob (ProductCategory c1 c2) = [ProductObject a b | a <- (ob c1), b <- (ob c2)]
        identity (ProductCategory c1 c2) (ProductObject a b) = ProductMorphism (identity c1 a) (identity c2 b)
        ar (ProductCategory c1 c2) (ProductObject s1 s2) (ProductObject t1 t2) = [ProductMorphism m1 m2 | m1 <- ar c1 s1 t1, m2 <- ar c2 s2 t2]
        
    instance (GeneratedFiniteCategory c1 m1 o1, Morphism m1 o1,
              GeneratedFiniteCategory c2 m2 o2, Morphism m2 o2) =>
              GeneratedFiniteCategory (ProductCategory c1 m1 o1 c2 m2 o2) (ProductMorphism m1 o1 m2 o2) (ProductObject o1 o2) where
        genAr (ProductCategory c1 c2) (ProductObject s1 s2) (ProductObject t1 t2) = [ProductMorphism m1 m2 | m1 <- genAr c1 s1 t1, m2 <- genAr c2 s2 t2]
        decompose (ProductCategory c1 c2) (ProductMorphism m1 m2) = [ProductMorphism d1 d2 | d1 <- decompoExtended1, d2 <- decompoExtended2]
            where
                decompo1 = decompose c1 m1
                decompo2 = decompose c2 m2
                decompoExtended1 = if ((length decompo1) < (length decompo2)) then (replicate ((length decompo2)-(length decompo1)) (identity c1 (target (head decompo1))))++decompo1 else decompo1
                decompoExtended2 = if ((length decompo2) < (length decompo1)) then (replicate ((length decompo1)-(length decompo2)) (identity c2 (target (head decompo2))))++decompo2 else decompo2
                
    instance (PrettyPrintable o1, PrettyPrintable o2) => PrettyPrintable (ProductObject o1 o2) where
        pprint (ProductObject a b) = "<"++(pprint a)++","++(pprint b)++">"
        
    
    instance (PrettyPrintable m1, PrettyPrintable m2) => PrettyPrintable (ProductMorphism m1 o1 m2 o2) where
        pprint (ProductMorphism f g) = "<"++(pprint f)++","++(pprint g)++">"